use crate::{err, partials, succ, AppState, Station};
use axum::{
    extract::State,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Form, Router,
};
use axum_extra::extract::{CookieJar, PrivateCookieJar};
use maud::{html, Markup, PreEscaped};
use serde::Deserialize;
use sqlx::SqlitePool;
use std::sync::Arc;
use tower_sessions::{cookie::Cookie, Session};

async fn station_picker(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    jar: CookieJar,
) -> Result<(CookieJar, Markup), (CookieJar, impl IntoResponse)> {
    let (jar, current_station_cookie) = get_station_cookie(&db, jar).await;
    if let Some(station) = current_station_cookie {
        return Err((jar, Redirect::to(&format!("/s/{}", station.id))));
    } else {
        let stations = Station::all(&db).await;
        let content = html! {
            h1 { "Wähle deine Station" }
            select onchange="window.location.href='/s/' + this.value;" {
                option selected value="" {
                    "Deine Station..."
                }
                @for station in stations {
                    option value=(station.id) { (station.name) };
                }
            }

        };
        Ok((jar, partials::page(content, session, false).await))
    }
}

async fn get_station_cookie(db: &SqlitePool, jar: CookieJar) -> (CookieJar, Option<Station>) {
    let Some(station_id) = jar.get("station_id") else {
        return (jar, None); // No station_id cookie
    };

    let station_id = match station_id.value().parse::<i64>() {
        Ok(number) => number,
        Err(_) => {
            // got some cookie which isn't a i64 -> destroy and start over again
            let jar = jar.remove(Cookie::from("station_id"));
            return (jar, None);
        }
    };

    let Some(station) = Station::find_by_id(db, station_id).await else {
        // got some cookie with an i64 which is no valid station_id -> destroy and start over again
        let jar = jar.remove(Cookie::from("station_id"));
        return (jar, None);
    };

    (jar, Some(station))
}

async fn decide_between_stations(
    cookie: &Station,
    trying_to_access: &Station,
    session: Session,
) -> Markup {
    let content = html! {
        h1 { "Wähle deine Station" }
        ul {
            li {
                "Die neu aufgerufene "
                (PreEscaped("&rarr; "))
                a href=(format!("/s/{}", trying_to_access.id)) {
                    button { (trying_to_access.name) }
                }
            }
            li {
                "Die Alte "
                (PreEscaped("&rarr; "))
                a href=(format!("/s/{}", cookie.id)) {
                    button { (cookie.name) }
                }
            }
        }
    };
    partials::page(content, session, false).await
}

async fn view(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    jar: CookieJar,
    pjar: PrivateCookieJar,
    axum::extract::Path(id): axum::extract::Path<i64>,
) -> Result<(CookieJar, PrivateCookieJar, Markup), (CookieJar, PrivateCookieJar, impl IntoResponse)>
{
    // Station selector
    let (mut jar, current_station_cookie) = get_station_cookie(&db, jar).await;
    if current_station_cookie.is_none() {
        jar = jar.add(Cookie::new("station_id", id.to_string()));
    }
    if let Some(current_station_cookie) = current_station_cookie {
        if current_station_cookie.id != id {
            // user has a cookie, which is a different station than she is trying to access
            if let Some(station) = Station::find_by_id(&db, id).await {
                jar = jar.remove(Cookie::from("station_id"));
                // trying to access valid station id
                err!(session, "Du hast versucht eine neue Station zu öffnen obwohl du bereits eine andere Station offen hattest. Welche möchtest du nun verwenden?");
                return Ok((
                    jar,
                    pjar,
                    decide_between_stations(&current_station_cookie, &station, session).await,
                ));
            } else {
                // user trying to access _in_valid station id -> make her aware + redirect to old
                err!(session, "Du hast versucht eine Station öffnen, die es nicht gibt. Nachdem du vorher schonmal eine andere Station (die es gibt) geöffnet hattest, bist du nun zu dieser weitergeleitet worden. Wenn du das nicht willst, logg dich bitte aus.");
                return Err((
                    jar,
                    pjar,
                    Redirect::to(&format!("/s/{}", current_station_cookie.id)),
                ));
            }
        }
    }
    let station = Station::find_by_id(&db, id).await.unwrap();

    let mut pjar = pjar;

    // PW Checker
    if let Some(pw) = pjar.get("pw") {
        if pw.value() != station.pw {
            pjar = pjar.remove(Cookie::from("station_id"));
            err!(session, "Du hattest einen falschen Code für Station {} gespeichert. Bitte gibt den richtigen ein:", station.name );
            return Err((jar, pjar, Redirect::to(&format!("/s/code",))));
        }
    } else {
        return Err((jar, pjar, Redirect::to(&format!("/s/code",))));
    }

    let content = html! {
        nav {
            ul {
                li { strong { (format!("Station {}", station.name)) } }
            }
            ul {
                li { a href="/s/station-logout" { "Logout" } }
            }
        }
        h1 { "test" }
    };

    Ok((jar, pjar, partials::page(content, session, false).await))
}

async fn code(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    jar: CookieJar,
    pjar: PrivateCookieJar,
) -> Result<(CookieJar, PrivateCookieJar, Markup), (CookieJar, PrivateCookieJar, impl IntoResponse)>
{
    let (jar, current_station_cookie) = get_station_cookie(&db, jar).await;
    let Some(station) = current_station_cookie else {
        return Err((jar, pjar, Redirect::to("/s")));
    };

    let content = html! {
        nav {
            ul {
                li { strong { (format!("Station {}", station.name)) } }
            }
            ul {
                li { a href="/s/station-logout" { "Station wechseln" } }
            }
        }
        h1 { "Code eingeben" }
        form action="/s/station-login" method="post" {
            fieldset role="group" {
                input type="text" name="pw" placeholder="Code" required;
                input type="submit" value="Login";
            }

        }
    };

    Ok((jar, pjar, partials::page(content, session, false).await))
}

#[derive(Deserialize)]
struct LoginForm {
    pw: String,
}

async fn login(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    jar: CookieJar,
    pjar: PrivateCookieJar,
    Form(form): Form<LoginForm>,
) -> (CookieJar, PrivateCookieJar, impl IntoResponse) {
    let (jar, current_station_cookie) = get_station_cookie(&db, jar).await;
    let Some(station) = current_station_cookie else {
        return (jar, pjar, Redirect::to("/s"));
    };

    let mut pjar = pjar;

    if form.pw == station.pw {
        pjar = pjar.add(Cookie::new("pw", form.pw));
        succ!(
            session,
            "Erfolgreich eingeloggt, viel Spaß beim bewerten :-)"
        );
        return (jar, pjar, Redirect::to(&format!("/s/{}", station.id)));
    } else {
        err!(session, "Falsches Passwort. Probiere es erneut.");
        std::thread::sleep(std::time::Duration::from_secs(1));
        return (jar, pjar, Redirect::to("/s/code"));
    }
}
async fn logout(session: Session, mut jar: CookieJar) -> (CookieJar, impl IntoResponse) {
    jar = jar.remove(Cookie::from("station_id"));

    succ!(session, "Erfolgreich ausgeloggt!");

    (jar, Redirect::to("/s"))
}

async fn quick_login(
    State(db): State<Arc<SqlitePool>>,
    session: Session,
    jar: CookieJar,
    pjar: PrivateCookieJar,
    axum::extract::Path((id, code)): axum::extract::Path<(i64, String)>,
) -> (CookieJar, PrivateCookieJar, impl IntoResponse) {
    if let Some(station) = Station::find_by_id(&db, id).await {
        if station.pw == code {
            succ!(
                session,
                "Erfolgreich eingeloggt, viel Spaß beim bewerten :-)"
            );
            let mut pjar = pjar.remove(Cookie::from("pw"));
            let mut jar = jar.remove(Cookie::from("station_id"));
            jar = jar.add(Cookie::new("station_id", id.to_string()));
            pjar = pjar.add(Cookie::new("pw", code));

            return (jar, pjar, Redirect::to(&format!("/s/{id}")));
        }
    }
    err!(
        session,
        "Falscher Quick-Einlogg-Link. Bitte nochmal scannen oder deine Station manuell auswählen:"
    );
    return (jar, pjar, Redirect::to("/s"));
}

pub(super) fn routes() -> Router<AppState> {
    Router::new()
        .route("/", get(station_picker))
        .route("/{id}", get(view))
        .route("/code", get(code))
        .route("/{id}/{code}", get(quick_login))
        .route("/station-login", post(login))
        .route("/station-logout", get(logout))
}
